#define __DEFINE_GFX_VARS
#include "manic.h"
#include "gfxlibs/keyboard.h"
#include "gfxlibs/gfx.h"
#include "gfxdata/gfxdata.h"
#include "miscdata/misc.h"
#include "functions.h"
#include "sndlibs/snd.h"
#include <string.h>


static void CheckCrumb (void);
static void CheckExit (void);
static void CheckKeys (void);
static void CheckSPG (void);
static void CheckSwitches (void);
static int CheckWillyConv (void);
static void CheckWillyFall (void);
static int CheckWillyKillBlock (void);
static void DoDeath (void);
static void DoEugene (void);
static void DoKongFall (void);
static void DoKongStood (void);
static void DoLevelSeven (void);
static void DoLevelThirteen (void);
static void DoSPG (void);
static void DoWillyFall (void);
static void DoWillyJump (int);
static void DoWillyLeft (void);
static void DoWillyRight (void);
static void DrawBGBlock (int xpos, int ypos);
static void DrawBlock (int xpos, int ypos, BYTE block, BYTE ink);
static void DrawConvBlock (int xpos, int ypos, BYTE block, BYTE frame);
static void DrawEUGENE (WORD xpos, WORD ypos, BYTE col);
static void DrawHRobo (WORD xpos, WORD ypos, WORD sprite, BYTE col);
static void DrawKONG (int xpos, int ypos, BYTE block, BYTE ink);
static void DrawKeyBlock (int xpos, int ypos, BYTE block, BYTE shine);
static void DrawSPGBlock (int x, int y);
static void DrawSky (int xpos, int ypos, BYTE block, BYTE ink);
static void DrawVROBO (WORD xpos, WORD ypos, WORD sprite, BYTE col);
static int GetBlock (WORD x, WORD y);
static int GetWillyInput (void);
static void PutWilly (void);
static int SPGCheckRobo (WORD x, WORD y);

BYTE bright[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 7, 6, 5, 4, 3, 2, 1 };
BYTE bright2[] = { 0, 1, 2, 3, 4, 3, 2, 1, 0, 1, 2, 3, 4, 3, 2, 1 };

static int delayBeep = 0;

/* ////////////////////////////////////////////////////////////
//	In-game music
//////////////////////////////////////////////////////////// */
void
StartGameMusic (void)
{
  mm_snd_playmod (MUSICtype ? 9 : (LEVEL == 19 || LEVEL == 39) ? 0x16 : 0);
  if (MUSICon)
    mm_snd_unmute ();
  else
    mm_snd_mute ();
}

/* ////////////////////////////////////////////////////////////
//	Print Score
//////////////////////////////////////////////////////////// */
void
PrintScore (void)
{
  INK = 6;
  PAPER = 0;
  FontPrint (20, 19, "Score");
  ShowSix (26, 19, SCORE);

  if (SCORE >= EXTRA) {
    LIVES++;
    EXTRA += EXTRAdelta;
    EXTRAc = 0;
    EXTRAm = 1;
  }
}

/* ////////////////////////////////////////////////////////////
//	Print High Score
//////////////////////////////////////////////////////////// */
void
PrintHighScore (void)
{
  INK = 6;
  PAPER = 0;
  FontPrint (0, 19, "High Score");
  ShowSix (11, 19, HISCORE);
}

/* ////////////////////////////////////////////////////////////
//	Redraw Stuff for Double Buffer
//////////////////////////////////////////////////////////// */
void
LevelDoneExtra (void)
{
  int i, count;
  BYTE temp;

  RemoveWilly ();
  RemoveSPG ();
  RemoveSky ();
  RemoveKong ();
  RemoveVRobo ();
  RemoveHRobo ();
  RemoveEugene ();
  RemoveSun ();

  if (cCONV.pos.y)
    for (count = 0; count < cCONV.l; count++)
      DrawConvBlock (cCONV.pos.x / 8 + count, cCONV.pos.y / 8, sCONV.gfx, cCONV.f);

  for (i = 0; i < 4; i++)
    if (cHROBO[i].pos.x < UNDEF)
      DrawHRobo ((cHROBO[i].pos.x & 248), cHROBO[i].pos.y,
		 (cHROBO[i].gfx + ((cHROBO[i].pos.x & cHROBO[i].anim) >> 1)) +
		 (cHROBO[i].d ? cHROBO[i].flip : 0), cHROBO[i].colour.ink);

  for (i = 0; i < 4; i++)
    if (cVROBO[i].pos.x < UNDEF)
      DrawVROBO ((cVROBO[i].pos.x & 248), cVROBO[i].pos.y,
		 cVROBO[i].gfx + ((cVROBO[i].pos.x & cVROBO[i].anim) >> 1),
		 cVROBO[i].colour.ink);

  if (LEVEL == 4)
    DrawEUGENE (EUGENE.pos.x, EUGENE.pos.y, EUGENE.c);

  if (LEVEL == 7 || LEVEL == 11)
    DrawKONG (KONG.pos.x, KONG.pos.y, KONG.f, KONG.c);

  if (LEVEL == 13)
    for (i = 0; i < 3; i++)
      DrawSky (SKY[i].pos.x, SKY[i].pos.y, SKY[i].f, SKY[i].c);

  if (LEVEL == 18)
    for (i = 0; i < 64; i++)
      if (SPG[old][i].x < UNDEF)
	DrawSPGBlock (SPG[old][i].x, SPG[old][i].y);

  for (i = 0; i < 5; i++)
    if (cKEY[i].s)
      if (cKEY[i].pos.x < UNDEF && cKEY[i].pos.y < UNDEF)
	DrawKeyBlock (cKEY[i].pos.x, cKEY[i].pos.y, cKEYgfx, bright2[cKEY[i].b]);

  PutWilly ();
  DrawExit ();
  DoSun ();

  temp = MUSICon;
  MUSICon = 0;
  DrawLives ();
  MUSICon = temp;
}

/* ////////////////////////////////////////////////////////////
//	Draw Lives
//////////////////////////////////////////////////////////// */
void
DrawLives (void)
{
  int i;
  BYTE l;

  l = (LIVES > 8) ? 8 : LIVES;

  for (i = 0; i < (l - 1); i++)
    Plot16 (i * 16, 168, GFXwilly + LIVESf * 256);

  if (CHEAT == 1)
    Plot16 (i * 16, 168, GFXover + 256); /* boot */

  if (MUSICon == 1) {
    LIVESp++;
    if (LIVESp > 3) {
      LIVESp = 0;
      LIVESf++;
      LIVESf &= 3;
    }
  }
}

/* ////////////////////////////////////////////////////////////
//	Draw Current Level
//////////////////////////////////////////////////////////// */
void
DrawCurrentLevel (void)
{
  int xpos, ypos;
  BYTE block;

  mm_gfx_cls (cBORDER);

  for (ypos = 0; ypos < 16; ypos++) {
    for (xpos = 0; xpos < 32; xpos++) {
      block = bMAP[(ypos * 32) + xpos];
      switch (block) {
      case 0:
	DrawBGBlock (xpos, ypos);
	break;
      case 1:
	DrawBlock (xpos, ypos, sPLAT1.gfx, sPLAT1.ink);
	break;
      case 2:
	DrawBlock (xpos, ypos, sPLAT2.gfx, sPLAT2.ink);
	break;
      case 3:
	DrawBlock (xpos, ypos, sWALL.gfx, sWALL.ink);
	break;
      case 4:
	DrawBlock (xpos, ypos, sCRUMB.gfx, sCRUMB.ink);
	break;
      case 5:
	DrawBlock (xpos, ypos, sKILL1.gfx, sKILL1.ink);
	break;
      case 6:
	DrawBlock (xpos, ypos, sKILL2.gfx, sKILL2.ink);
	break;
      }
    }
  }
  if (LEVEL == 19) {
    DrawFinal ();
    SUN.m = 1;
    SUN.y = 32;
    SUN.h = 16;
    DoSun ();
  }
}

/* ////////////////////////////////////////////////////////////
//	Draw a BackGround Block
//////////////////////////////////////////////////////////// */
static void
DrawBGBlock (int xpos, int ypos)
{
  Plot8back (xpos, ypos, cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Draw a Block
//////////////////////////////////////////////////////////// */
static void
DrawBlock (int xpos, int ypos, BYTE block, BYTE ink)
{
  Plot8ink (xpos, ypos, GFXblocks + block * 64, ink, cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Draw a Conveyor Block
//////////////////////////////////////////////////////////// */
static void
DrawConvBlock (int xpos, int ypos, BYTE block, BYTE frame)
{
  Plot8ink (xpos, ypos, GFXconv + block * 256 + frame * 64, sCONV.ink,
	    cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Draw Conveyor
//////////////////////////////////////////////////////////// */
void
DrawConv (void)
{
  int count;

  if (cCONV.pos.y != 0) {
    for (count = 0; count < cCONV.l; count++)
      DrawConvBlock (cCONV.pos.x / 8 + count, cCONV.pos.y / 8, sCONV.gfx, cCONV.f);
    switch (cCONV.d) {
    case 0:
      cCONV.f++;
      break;
    case 1:
      cCONV.f--;
      break;
    }
    cCONV.f &= 3;
  }
}

/* ////////////////////////////////////////////////////////////
//	Draw Exit
//////////////////////////////////////////////////////////// */
void
DrawExit (void)
{
  BYTE *block = GFXexit + cEXIT.gfx * 256;

  switch (cEXIT.m) {
  case 0:
    Plot16mask (cEXIT.pos.x, cEXIT.pos.y, block);
    break;

  case 1:
    PlotShiny (cEXIT.pos.x, cEXIT.pos.y, block, bright[cEXIT.b], 16);
    cEXIT.b = (cEXIT.b + 1) & 15;
  }
}

/* ////////////////////////////////////////////////////////////
//	Draw a Key
//////////////////////////////////////////////////////////// */
static void
DrawKeyBlock (int xpos, int ypos, BYTE block, BYTE shine)
{
  PlotShiny (xpos, ypos, GFXkey + block * 64, shine, 8);
}

/* ////////////////////////////////////////////////////////////
//	Draw Keys
//////////////////////////////////////////////////////////// */
void
DrawKeys (void)
{
  int i, count;

  count = 0;

  for (i = 0; i < 5; i++) {
    if (cKEY[i].s) {
      if (cKEY[i].pos.x < UNDEF && cKEY[i].pos.y < UNDEF) {
	DrawKeyBlock (cKEY[i].pos.x, cKEY[i].pos.y, cKEYgfx, bright2[cKEY[i].b]);
	count++;
      }
      cKEY[i].b++;
      cKEY[i].b &= 15;
    }
    else
      if (cKEY[i].pos.x < UNDEF && cKEY[i].pos.y < UNDEF)
	DrawBGBlock (cKEY[i].pos.x / 8, cKEY[i].pos.y / 8);
  }

  if (!count)
    cEXIT.m = 1;
}

/* ////////////////////////////////////////////////////////////
//	Draw Level Title Plate
//////////////////////////////////////////////////////////// */
void
DrawTPlate (void)
{
  PlotXY (0, 128, GFXtplate, 256, 8);
  FontPrint2 (0, 16, bTITLE);
}

/* ////////////////////////////////////////////////////////////
//	Draw Air
//////////////////////////////////////////////////////////// */
void
DrawAir (void)
{
  int x;

  if (!--cAIRp) {
    cAIRp = 8;
    if (--cAIR <= 33) {
      cAIR = 33;
      cWILLY.m = 6;
    }
  }

  for (x = 3; x >= 0; --x)
    PlotXY (30, 138 + x, GFXair + 512 + 256 * x + 30, 226, 1);

  mm_gfx_putpixel (32, 138, 122);
  mm_gfx_putpixel (32, 139, 116);
  mm_gfx_putpixel (32, 140, 119);
  mm_gfx_putpixel (32, 141, 123);
  mm_gfx_setdirty (32, 138, 1, 4);

  if (cAIR > 33) {
    PlotXYback (33, 138, 120, cAIR - 33, 1);
    PlotXYback (33, 139, 114, cAIR - 33, 1);
    PlotXYback (33, 140, 117, cAIR - 33, 1);
    PlotXYback (33, 141, 121, cAIR - 33, 1);
  }

  if (cAIR > 32) {
    mm_gfx_putpixel (cAIR - 1, 138, 122);
    mm_gfx_putpixel (cAIR - 1, 139, 116);
    mm_gfx_putpixel (cAIR - 1, 140, 119);
    mm_gfx_putpixel (cAIR - 1, 141, 123);
    mm_gfx_setdirty (cAIR - 1, 138, 1, 4);
  }
}

/* ////////////////////////////////////////////////////////////
//	Draw a HORIZONTAL Robot
//////////////////////////////////////////////////////////// */
static void
DrawHRobo (WORD xpos, WORD ypos, WORD sprite, BYTE col)
{
  Plot16mask_ink (xpos, ypos, GFXhrobo + sprite * 256, col - 1);
}

/* ////////////////////////////////////////////////////////////
//	Do Horizontal Robots
//////////////////////////////////////////////////////////// */
void
DoHRobo (void)
{
  int i;

  for (i = 0; i < 4; i++) {
    if (cHROBO[i].pos.x < UNDEF) {
      if (cHROBO[i].d) {
	cHROBO[i].pos.x -= (2 >> cHROBO[i].s);
	if (cHROBO[i].pos.x < cHROBO[i].min) {
	  cHROBO[i].d = 0;
	  cHROBO[i].pos.x += (2 >> cHROBO[i].s);
	}
      }
      else {
	cHROBO[i].pos.x += (2 >> cHROBO[i].s);
	if (cHROBO[i].pos.x > (cHROBO[i].max + 6)) {
	  cHROBO[i].d = 1;
	  cHROBO[i].pos.x -= (2 >> cHROBO[i].s);
	}
      }
      DrawHRobo ((cHROBO[i].pos.x & 248), cHROBO[i].pos.y,
		 (cHROBO[i].gfx + ((cHROBO[i].pos.x & cHROBO[i].anim) >> 1)) +
		 (cHROBO[i].d ? cHROBO[i].flip : 0), cHROBO[i].colour.ink);
      cHROBO[i].old[old].x = cHROBO[i].pos.x & 248;
      cHROBO[i].old[old].y = cHROBO[i].pos.y;
    }
  }
}

/* ////////////////////////////////////////////////////////////
//	Remove Horizontal Robots
//////////////////////////////////////////////////////////// */
void
RemoveHRobo (void)
{
  int i;
  for (i = 3; i >= 0; --i)
    if (cHROBO[i].old[old].x < UNDEF)
      Plot16back (cHROBO[i].old[old].x, cHROBO[i].old[old].y, cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Draw a VERTICAL Robot
//////////////////////////////////////////////////////////// */
static void
DrawVROBO (WORD xpos, WORD ypos, WORD sprite, BYTE col)
{
  Plot16mask_ink (xpos, ypos, GFXvrobo + sprite * 256, col - 1);
}

/* ////////////////////////////////////////////////////////////
//	Do Vertical Robots
//////////////////////////////////////////////////////////// */
void
DoVRobo (void)
{
  int i;
  for (i = 0; i < 4; i++) {
    if (cVROBO[i].pos.x < UNDEF) {
      if (cVROBO[i].d) {
	cVROBO[i].pos.y -= (cVROBO[i].s);
	if ((cVROBO[i].pos.y < cVROBO[i].min) || (cVROBO[i].pos.y > 256)) {
	  cVROBO[i].d = 0;
	  cVROBO[i].pos.y += (cVROBO[i].s);
	}
      }
      else {
	cVROBO[i].pos.y += (cVROBO[i].s);
	if (cVROBO[i].pos.y > cVROBO[i].max) {
	  cVROBO[i].d = 1;
	  cVROBO[i].pos.y -= (cVROBO[i].s);
	}
      }
      DrawVROBO (cVROBO[i].pos.x, cVROBO[i].pos.y, (cVROBO[i].gfx + cVROBO[i].anim),
		 cVROBO[i].colour.ink);
      cVROBO[i].anim++;
      cVROBO[i].anim &= 3;
      cVROBO[i].old[old].x = cVROBO[i].pos.x;
      cVROBO[i].old[old].y = cVROBO[i].pos.y;
    }
  }
}

/* ////////////////////////////////////////////////////////////
//	Remove Vertical Robots
//////////////////////////////////////////////////////////// */
void
RemoveVRobo (void)
{
  int i;
  for (i = 3; i >= 0; --i)
    if (cVROBO[i].old[old].x < UNDEF)
      Plot16back (cVROBO[i].old[old].x, cVROBO[i].old[old].y, cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Draw Switches
//////////////////////////////////////////////////////////// */
void
DrawSwitches (void)
{
  int i;
  for (i = 0; i < 2; i++)
    if (cSWITCH[i].s && cSWITCH[i].pos.x < UNDEF)
      Plot8ink (cSWITCH[i].pos.x / 8, cSWITCH[i].pos.y / 8,
		GFXswitch + (cSWITCH[i].s - 1) * 64, sPLAT2.ink, cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Do Special Robo
//////////////////////////////////////////////////////////// */
void
DoSpecialRobo (void)
{
  switch (LEVEL) {
  case 4:
    DoEugene ();
    break;
  case 7:
    DoLevelSeven ();
    break;
  case 11:
    DoLevelSeven ();
    break;
  case 13:
    DoLevelThirteen ();
    break;
  case 18:
    DoSPG ();
    CheckSPG ();
    break;
  }
}

/* ////////////////////////////////////////////////////////////
//	Draw Eugene
//////////////////////////////////////////////////////////// */
static void
DrawEUGENE (WORD xpos, WORD ypos, BYTE col)
{
  Plot16mask_ink (xpos, ypos, GFXeugene, col - 1);
}

/* ////////////////////////////////////////////////////////////
//	Remove Eugene
//////////////////////////////////////////////////////////// */
void
RemoveEugene (void)
{
  if (EUGENE.old[old].x < UNDEF)
    Plot16back (EUGENE.pos.x, EUGENE.pos.y, cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Do Eugene
//////////////////////////////////////////////////////////// */
static void
DoEugene (void)
{
  if (cEXIT.m == 1)
    EUGENE.m = 1;

  switch (EUGENE.m) {
  case 0:
    if (EUGENE.d) {
      if (--EUGENE.pos.y < EUGENE.min) {
	EUGENE.d = 0;
	EUGENE.pos.y++;
      }
    }
    else {
      if (++EUGENE.pos.y > EUGENE.max) {
	EUGENE.d = 1;
	EUGENE.pos.y--;
      }
    }
    DrawEUGENE (EUGENE.pos.x, EUGENE.pos.y, EUGENE.c);
    EUGENE.old[old].x = EUGENE.pos.x;
    EUGENE.old[old].y = EUGENE.pos.y;
    break;

  case 1:
    EUGENE.c = (EUGENE.c & 7) + 1;
    if (EUGENE.pos.y < EUGENE.max)
      EUGENE.pos.y++;
    DrawEUGENE (EUGENE.pos.x, EUGENE.pos.y, EUGENE.c);
    EUGENE.old[old].x = EUGENE.pos.x;
    EUGENE.old[old].y = EUGENE.pos.y;
  }
}

/* ////////////////////////////////////////////////////////////
//	Do Hole
//////////////////////////////////////////////////////////// */
static void
DoHole (void)
{
  switch (SWITCHm[0]) {
  case 0:
    PlotXYback (136, HOLEy, cBG.paper, 8, HOLEl);
    HOLEy--;
    HOLEl += 2;
    if (HOLEy == 87) {
      SWITCHm[0] = 1;
      bMAP[((88 / 8) * 32) + (136 / 8)] = 0;
      bMAP[((96 / 8) * 32) + (136 / 8)] = 0;
    }
    break;

  case 1:
    PlotXYback (136, 88, cBG.paper, 8, 16);
    SWITCHm[0] = 2;
    cHROBO[1].max += 24;
    break;
  }
}

/* ////////////////////////////////////////////////////////////
//	Level 7 "Miner Willy meets the Kong Beast"
//////////////////////////////////////////////////////////// */
static void
DoLevelSeven (void)
{
  if (--KONG.p == 0) {
    KONG.p = KONGPAUSE;
    KONG.f++;
    KONG.f &= 1;
  }

  if (cSWITCH[0].s == 2)
    DoHole ();

  if (cSWITCH[1].s == 1)
    DoKongStood ();
  else if (cSWITCH[1].s == 2)
    DoKongFall ();
}

/* ////////////////////////////////////////////////////////////
//	Do Kong Stood
//////////////////////////////////////////////////////////// */
static void
DoKongStood (void)
{
  DrawKONG (KONG.pos.x, KONG.pos.y, KONG.f, KONG.c);

  KONG.old[old].x = KONG.pos.x;
  KONG.old[old].y = KONG.pos.y;
}

/* ////////////////////////////////////////////////////////////
//	Do Kong Falling
//////////////////////////////////////////////////////////// */
static void
DoKongFall (void)
{
  switch (KONG.m) {
  case 0:
    KONG.m = 1;

    bMAP[((16 / 8) * 32) + (120 / 8)] = 0;
    bMAP[((16 / 8) * 32) + (128 / 8)] = 0;

    DrawBGBlock (120 / 8, 16 / 8);
    DrawBGBlock (128 / 8, 16 / 8);

    DrawKONG (KONG.pos.x, KONG.pos.y, KONG.f, KONG.c);
    KONG.old[old].x = KONG.pos.x;
    KONG.old[old].y = KONG.pos.y;
    break;

  case 1:
    KONG.m = 2;

    DrawBGBlock (120 / 8, 16 / 8);
    DrawBGBlock (128 / 8, 16 / 8);

    DrawKONG (KONG.pos.x, KONG.pos.y, KONG.f, KONG.c);
    KONG.old[old].x = KONG.pos.x;
    KONG.old[old].y = KONG.pos.y;

    KONG.c = 5;
    break;

  case 2:
    KONG.pos.y += 4;

    mm_snd_jump (22050 - (KONG.pos.y * 50));

    DrawKONG (KONG.pos.x, KONG.pos.y, KONG.f + 2, KONG.c);
    if (KONG.pos.y >= KONG.max)
      KONG.m = 3;
    else
      SCORE += 100;
    KONG.old[old].x = KONG.pos.x;
    KONG.old[old].y = KONG.pos.y;
    break;
  }
}

/* ////////////////////////////////////////////////////////////
//	Draw Kong
//////////////////////////////////////////////////////////// */
static void
DrawKONG (int xpos, int ypos, BYTE block, BYTE ink)
{
  Plot16mask_ink (xpos, ypos, GFXkong + 256 * block, ink);
}

/* ////////////////////////////////////////////////////////////
//	Remove Kong
//////////////////////////////////////////////////////////// */
void
RemoveKong (void)
{
  if (KONG.old[old].x < UNDEF)
    Plot16back (KONG.old[old].x, KONG.old[old].y, cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Draw Skylab
//////////////////////////////////////////////////////////// */
static void
DrawSky (int xpos, int ypos, BYTE block, BYTE ink)
{
  Plot16mask_ink (xpos, ypos, GFXsky + 256 * block, ink);
}

/* ////////////////////////////////////////////////////////////
//	Remove Skylab
//////////////////////////////////////////////////////////// */
void
RemoveSky (void)
{
  int i;

  for (i = 0; i < 3; i++)
    if (SKY[i].old[old].x < UNDEF)
      Plot16back (SKY[i].old[old].x, SKY[i].old[old].y, cBG.paper);
}

/* ////////////////////////////////////////////////////////////
//	Level 13 "       Skylab Landing Bay       "
//////////////////////////////////////////////////////////// */
static void
DoLevelThirteen (void)
{
  int i;

  for (i = 0; i < 3; i++) {
    switch (SKY[i].m) {
    case 0:
      SKY[i].pos.y += SKY[i].s;
      if (SKY[i].pos.y > SKY[i].max) {
	SKY[i].pos.y = SKY[i].max;
	SKY[i].m = 1;
	SKY[i].f++;
      }
      DrawSky (SKY[i].pos.x, SKY[i].pos.y, SKY[i].f, SKY[i].c);
      break;

    case 1:
      SKY[i].f++;

      if (SKY[i].f == 7) {
	SKY[i].m = 2;
      }
      DrawSky (SKY[i].pos.x, SKY[i].pos.y, SKY[i].f, SKY[i].c);
      break;

    case 2:
      SKY[i].p++;
      SKY[i].p &= 3;

      SKY[i].pos.x = SKYp[i][SKY[i].p].x;
      SKY[i].pos.y = SKYp[i][SKY[i].p].y;

      SKY[i].f = 0;
      SKY[i].m = 0;
    }

    SKY[i].old[old].x = SKY[i].pos.x;
    SKY[i].old[old].y = SKY[i].pos.y;
  }
}

/* ////////////////////////////////////////////////////////////
//	Draw a SPG Block
//////////////////////////////////////////////////////////// */
static void
DrawSPGBlock (int x, int y)
{
  int c1, c2;
  BYTE data;

  x *= 8;
  y *= 8;

  for (c1 = 7; c1 >= 0; --c1) {
    for (c2 = 7; c2 >= 0; --c2) {
      data = mm_gfx_getpixel (x + c2, y + c1);
      data = (data & 15) - 4;
      if (data > 15)
	data = 0;
      mm_gfx_putpixel (x + c2, y + c1, data + 96);
    }
  }
  mm_gfx_setdirty (x, y, 7, 7);
}

/* ////////////////////////////////////////////////////////////
//	Find Route
//////////////////////////////////////////////////////////// */
static void
FindSPG (void)
{
  int x = 23, y = 0, i, done = 0;
  BYTE dir = 0;

  for (i = 0; i < 64; i++)
    SPG[old][i].x = SPG[old][i].y = UNDEF;

  i = 0;
  do {
    int blockhit = bMAP[(y * 32) + x];
    int robohit = SPGCheckRobo (x, y);

    if (blockhit && robohit) {
      SPG[old][i].x = SPG[old][i].y = UNDEF;
      done = 1;
    }

    else if (!blockhit && robohit) {
      if (SPG[old][i - 1].x == x && SPG[old][i - 1].y == y) {
	SPG[old][i].x = SPG[old][i].y = UNDEF;
	done = 1;
      }
      SPG[old][i].x = x;
      SPG[old][i].y = y;
      i++;
      dir ^= 1;
    }

    else if (!blockhit && !robohit) {
      SPG[old][i].x = x;
      SPG[old][i].y = y;
      i++;
    }

    else if (blockhit && !robohit) {
      if (SPG[old][i - 1].x == x && SPG[old][i - 1].y == y) {
	SPG[old][i].x = SPG[old][i].y = UNDEF;
	done = 1;
      }
      dir ^= 1;
    }

    if (!blockhit) {
      if (!dir) {
	blockhit = bMAP[(++y * 32) + x];
	if (y == 15 || blockhit) {
	  SPG[old][i].x = SPG[old][i].y = UNDEF;
	  done = 1;
	}
      }
      else {
	blockhit = bMAP[(y * 32) + --x];
	if (x == 0 || blockhit) {
	  SPG[old][i].x = SPG[old][i].y = UNDEF;
	  done = 1;
	}
      }
    }
    else {
      if (!dir) {
	if (!--x) {
	  SPG[old][i].x = SPG[old][i].y = UNDEF;
	  done = 1;
	}
      }
      else {
	if (++y == 15) {
	  SPG[old][i].x = SPG[old][i].y = UNDEF;
	  done = 1;
	}
      }
    }

  } while (!done);

  SPG[old][i].x = SPG[old][i].y = UNDEF;
}

/* ////////////////////////////////////////////////////////////
//	Check if hit Robo
//////////////////////////////////////////////////////////// */
static int
SPGCheckRobo (WORD x, WORD y)
{
  int i;
  int hit;

  x *= 8;
  y *= 8;

  hit = 0;

  for (i = 0; i < 4; i++) {
    if (cHROBO[i].pos.x < UNDEF) {
      if ((x + 7) >= cHROBO[i].pos.x && x <= (cHROBO[i].pos.x + 15) &&
	  (y + 7) >= cHROBO[i].pos.y && y <= (cHROBO[i].pos.y + 15)) {
	hit++;
      }
    }
    if (cVROBO[i].pos.x < UNDEF) {
      if ((x + 7) >= cVROBO[i].pos.x && x <= (cVROBO[i].pos.x + 15) &&
	  (y + 7) >= cVROBO[i].pos.y && y <= (cVROBO[i].pos.y + 15)) {
	hit++;
      }
    }
  }
  return hit;
}

/* ////////////////////////////////////////////////////////////
//	Do the Solar Power Generator
//////////////////////////////////////////////////////////// */
static void
DoSPG (void)
{
  int i;

  FindSPG ();
  for (i = 0; i < 64; i++)
    if (SPG[old][i].x < UNDEF)
      DrawSPGBlock (SPG[old][i].x, SPG[old][i].y);
}

/* ////////////////////////////////////////////////////////////
//	Check SPG
//////////////////////////////////////////////////////////// */
static void
CheckSPG (void)
{
  int i;

  for (i = 0; i < 64 && SPG[old][i].x < UNDEF; i++)
    if (SPG[old][i].x * 8 + 7 >= cWILLY.pos.x && SPG[old][i].x * 8 < cWILLY.pos.x + 8 &&
	SPG[old][i].y * 8 + 7 >= cWILLY.pos.y && SPG[old][i].y * 8 < cWILLY.pos.y + 16) {
      cAIR--;
      return;
    }
}

/* ////////////////////////////////////////////////////////////
//	Draw Willy
//////////////////////////////////////////////////////////// */
static int
DrawWilly (int xpos, int ypos, BYTE block)
{
  int x, y;
  BYTE *block2, data;

  block2 = GFXwilly + 256 * block + 256;

  for (y = 15; y >= 0; --y)
    for (x = 15; x >= 0; --x) {
      data = *--block2;
      if (data)
	if (mm_gfx_getpixel (xpos + x, ypos + y) == cBG.paper)
	  mm_gfx_putpixel (xpos + x, ypos + y, data);
    }
  mm_gfx_setdirty (xpos, ypos, 15, 15);

  return 0;
}

/* ////////////////////////////////////////////////////////////
//	Get Willy Background
//////////////////////////////////////////////////////////// */
static void
GetWillyBG (int xpos, int ypos)
{
  int x, y;

  for (y = 15; y >= 0; --y)
    for (x = 15; x >= 0; --x)
      cWILLY.buff[old][(y * 16) + x] = mm_gfx_getpixel (xpos + x, ypos + y);
}

/* ////////////////////////////////////////////////////////////
//	Remove Willy
//////////////////////////////////////////////////////////// */
void
RemoveWilly (void)
{
  Plot16 (cWILLY.old[old].x, cWILLY.old[old].y, cWILLY.buff[old]);
}

/* ////////////////////////////////////////////////////////////
//  Check if hit Robo
//////////////////////////////////////////////////////////// */
typedef BYTE frame_t[16];

static int
doPixelCheck (int rx, int ry, const BYTE * robo)
{
  static BYTE grid[16][16];
  frame_t *wptr = (frame_t *) robo;
  int x, y, w;

  rx -= cWILLY.pos.x & 248;
  ry -= cWILLY.pos.y;
  if (rx < -15 || rx > 15 || ry < -15 || ry > 15)
    return 0;
  memset (grid, 0, sizeof (grid));
  if (rx < 0) {
    x = 0;
    rx = -rx;
    w = 16 - rx;
  }
  else {
    x = rx;
    rx = 0;
    w = 16 - x;
  }
  for (y = ry + 15; y >= ry; y--)	/* partial plot robo */
    if (y >= 0 && y < 16)
      memcpy (&grid[y][x], &wptr[y - ry][rx], w);
  wptr =
    (frame_t *) (GFXwilly + ((cWILLY.pos.x & 15) >> 1) * 256 +
		 (cWILLY.d ? 8 * 256 : 0));
  for (x = 255; x >= 0; x--)	/* check for collision */
    if (wptr[x >> 4][x & 15] && grid[x >> 4][x & 15])
      return 1;
  return 0;
}

static int
WillyCheckRobo (void)
{
  int i;
  int hit = 0;

  for (i = 3; i >= 0; i--) {
    if (cHROBO[i].pos.x < UNDEF)
      hit += doPixelCheck (cHROBO[i].pos.x & 248, cHROBO[i].pos.y,
			   GFXhrobo + 256 * (cHROBO[i].gfx +
					     ((cHROBO[i].pos.x & cHROBO[i].anim) >>
					      1) +
					     (cHROBO[i].d ? cHROBO[i].flip :
					      0)));
    if (cVROBO[i].pos.x < UNDEF)
      hit +=
	doPixelCheck (cVROBO[i].pos.x & 248, cVROBO[i].pos.y,
		      GFXvrobo + 256 * (cVROBO[i].gfx + cVROBO[i].anim));
  }

  if (LEVEL == 4 || LEVEL == 24)
    hit += doPixelCheck (EUGENE.pos.x, EUGENE.pos.y, GFXeugene + (TONKS ? 256 : 0));

  if (KONG.m != 3 && (LEVEL == 7 || LEVEL == 11 || LEVEL == 27 || LEVEL == 31))
    hit += doPixelCheck (KONG.pos.x, KONG.pos.y, GFXkong + 256 * KONG.f);

  if (LEVEL == 13 || LEVEL == 33)
    for (i = 2; i >= 0; i--)
      hit += doPixelCheck (SKY[i].pos.x, SKY[i].pos.y, GFXsky + 256 * SKY[i].f);

  return hit;
}

/* ////////////////////////////////////////////////////////////
//	Do Willy Stuff
//
//	Modes:-
//	0=Walking Left\Right
//	1=Jump Up
//	2=Jump Left
//	3=jump Right
//	4=Fall
//	5=On Conv
//	6=Death
//
//////////////////////////////////////////////////////////// */
void
DoWilly (void)
{
  int input, onconv;
  if (!PAUSE) {
    input = GetWillyInput ();
    onconv = CheckWillyConv ();

    if (CheckWillyKillBlock ())
      cWILLY.m = 6;
    if (cWILLY.m == 0 && (cWILLY.j != 1 || (input & 4) == 0))
      CheckWillyFall ();

    if (cWILLY.m < 1 || 4 < cWILLY.m)
      delayBeep = 0;

    switch (cWILLY.m) {
    case 0:
      CheckCrumb ();
      switch (input) {
      case 1:
	cWILLY.j = 1;
	cWILLY.fall = 0;
	if (onconv && cCONV.d == 1) {
	  if (cWILLY.d && cCONV.m) {
	    if (cCONV.m == 1)
	      DoWillyLeft ();
	  }
	  else
	    DoWillyRight ();
	}
	else {
	  cCONV.m = 1;
	  DoWillyLeft ();
	}
	break;
      case 2:
	cWILLY.j = 1;
	cWILLY.fall = 0;
	if (onconv && cCONV.d == 0) {
	  if (cWILLY.d == 0 && cCONV.m) {
	    if (cCONV.m == 1)
	      DoWillyRight ();
	  }
	  else
	    DoWillyLeft ();
	}
	else {
	  cCONV.m = 1;
	  DoWillyRight ();
	}
	break;
      case 4:
	if (cWILLY.j > 1)
	  cWILLY.j = 0;
	cWILLY.m = 1;
	cWILLY.fall = 0;
	cWILLY.jp = 0;
	if (onconv) {
	  if (cWILLY.d != cCONV.d) {	/* conv, same dir... */
	    cWILLY.m = 3 - cWILLY.d;
	  }
	  else {
	    cWILLY.d ^= 1;
	  }
	}
	goto doJump;
      case 5:
	if (cWILLY.j > 1)
	  cWILLY.j = 0;
	cWILLY.fall = 0;
	cWILLY.jp = 0;
	if (!onconv) {		/* not on conveyor */
	  if (cWILLY.d) {	/* left */
	    cWILLY.m = 2;
	  }
	  else {
	    cWILLY.m = 1;
	    cWILLY.d = 1;
	  }
	}
	else {			/* on conveyor */
	  if (cWILLY.d != cCONV.d) {	/* same dir */
	    cWILLY.m = cWILLY.d ? 2 :
	      (GetBlock (cWILLY.pos.x, cWILLY.pos.y - 1) != 3 &&
	       GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y - 1) != 3) ? 3 : 0;
	  }
	  else {		/* opposite dirs */
	    if (cCONV.m || cWILLY.d) {
	      if (cCONV.m &&
		  (GetBlock (cWILLY.pos.x, cWILLY.pos.y - 1) == 3 ||
		   GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y - 1) == 3))
		cCONV.m = 0;
	      cWILLY.m = cCONV.m == 1 ? 2 : 1;
	    }
	    else {
	      cWILLY.m = 1;
	      cWILLY.d ^= 1;	/* still: turn and jump up */
	    }
	  }
	}
	if (cWILLY.m)
	  goto doJump;
	else
	  break;
      case 6:
	if (cWILLY.j > 1)
	  cWILLY.j = 0;
	cWILLY.fall = 0;
	cWILLY.jp = 0;
	if (!onconv) {		/* not on conveyor */
	  if (cWILLY.d == 0) {	/* right */
	    cWILLY.m = 3;
	  }
	  else {
	    cWILLY.m = 1;
	    cWILLY.d = 0;
	  }
	}
	else {			/* on conveyor */
	  if (cWILLY.d != cCONV.d) {	/* same dir */
	    cWILLY.m = !cWILLY.d ? 3 :
	      (GetBlock (cWILLY.pos.x, cWILLY.pos.y - 1) != 3 &&
	       GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y - 1) != 3) ? 2 : 0;
	  }
	  else {		/* opposite dirs */
	    if (cCONV.m || cWILLY.d == 0) {
	      if (cCONV.m &&
		  (GetBlock (cWILLY.pos.x, cWILLY.pos.y - 1) == 3 ||
		   GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y - 1) == 3))
		cCONV.m = 0;
	      cWILLY.m = cCONV.m == 1 ? 3 : 1;
	    }
	    else {
	      cWILLY.m = 1;
	      cWILLY.d ^= 1;	/* still: turn and jump up */
	    }
	  }
	}
	if (cWILLY.m == 0)
	  break;
      doJump:DoWillyJump (cWILLY.m == 1 ? 0 : cWILLY.m == 2 ? -1 : 1);
	break;
      default:
	cWILLY.j = 0;
	cWILLY.fall = 0;
	cWILLY.jp = 0;
	if (onconv) {
	  if (cCONV.d)
	    DoWillyRight ();
	  else
	    DoWillyLeft ();
	}
	else
	  cCONV.m = 0;
      }
      break;
    case 1:
      DoWillyJump (0);
      break;
    case 2:
      DoWillyJump (-1);
      break;
    case 3:
      DoWillyJump (1);
      break;
    case 4:
      DoWillyFall ();
      break;
    case 6:
      DoDeath ();
      break;
    }
    if (WillyCheckRobo ())
      cWILLY.m = 6;
    else {
      CheckKeys ();
      CheckExit ();
      CheckSwitches ();
    }
  }
  PutWilly ();
  cWILLY.old[old].x = cWILLY.pos.x & 248;
  cWILLY.old[old].y = cWILLY.pos.y;
}

/* ////////////////////////////////////////////////////////////
//	Do Willy Left
//////////////////////////////////////////////////////////// */
static void
DoWillyLeft (void)
{
  int block1, block2, block3;

  if (cWILLY.d == 0)
    cWILLY.d = 1;
  else {
    cWILLY.pos.x -= 2;
    block1 = GetBlock (cWILLY.pos.x, cWILLY.pos.y);
    block2 = GetBlock (cWILLY.pos.x, cWILLY.pos.y + 8);
    block3 = (cWILLY.pos.y & 7) ? GetBlock (cWILLY.pos.x, cWILLY.pos.y + 16) : block2;
    if (block1 == 3 || block2 == 3 || block3 == 3)
      cWILLY.pos.x += 2;
  }
}

/* ////////////////////////////////////////////////////////////
//	Do Willy Right
//////////////////////////////////////////////////////////// */
static void
DoWillyRight (void)
{
  int block1, block2, block3;

  if (cWILLY.d == 1)
    cWILLY.d = 0;
  else {
    cWILLY.pos.x += 2;
    block1 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y);
    block2 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y + 8);
    block3 = (cWILLY.pos.y & 7) ? GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y + 16) : block2;
    if (block1 == 3 || block2 == 3 || block3 == 3)
      cWILLY.pos.x -= 2;
  }
}

/* ////////////////////////////////////////////////////////////
//	Do Willy Jump
//////////////////////////////////////////////////////////// */
static int cWILLYjp[] =
  { 0, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4 };
  /* ends in fall */

static void
DoWillyJump (int dir)
{
  int block1, block2;

  cWILLY.pos.y += cWILLYjp[cWILLY.j];
  if (cWILLY.j < 9) {
    if (cWILLY.j &&
	(GetBlock (cWILLY.pos.x, cWILLY.pos.y) == 3
	 || GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y) == 3)) {
      cWILLY.m = 4;
      cWILLY.fall = -4;
      cCONV.m = 0;
      delayBeep = cWILLY.j > 1;
      DoWillyFall ();
    }
    else {
      if (dir < 0)
	DoWillyLeft ();
      else if (dir > 0)
	DoWillyRight ();
    }
  }
  else {
    int x = cWILLY.pos.x;
    if (cWILLY.j > 12)
      cWILLY.fall += cWILLYjp[cWILLY.j];
    if (cWILLY.j < 18) {
      if (dir < 0)
	DoWillyLeft ();
      else if (dir > 0)
	DoWillyRight ();
    }
    if (cWILLY.j > 10 && (cWILLY.pos.y & 7) == 0) {
      block1 = GetBlock (x, cWILLY.pos.y + 16);
      block2 = GetBlock (x + 8, cWILLY.pos.y + 16);
      if (block1 || block2) {
	cWILLY.pos.x = x;
	cWILLY.m = 0;
	if (block1 == 7 || block2 == 7) {
	  cCONV.m = dir != 0;
	  cWILLY.fall = 0;
	}
      }
    }
  }

  if (++cWILLY.j == 19)
    cWILLY.m = 0;
  if (cWILLY.j < 11)
    cWILLY.jp = -cWILLY.j;
  else
    cWILLY.jp++;

  if (cWILLY.pos.y > 104) {
    /*cWILLY.pos.y=104; */
    cWILLY.m = 4;
    cCONV.m = dir;
    DoWillyFall ();
  }

  if (delayBeep)
    mm_snd_jump (16384 - cWILLY.jp * 1500);
  else
    delayBeep=1;
}

/* ////////////////////////////////////////////////////////////
//	Check Fall
//////////////////////////////////////////////////////////// */
static void
CheckWillyFall (void)
{
  int block1, block2;

  if (cWILLY.pos.y & 7)
    block1 = block2 = 0;
  else {
    block1 = GetBlock (cWILLY.pos.x, cWILLY.pos.y + 16);
    block2 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y + 16);
  }
  if (block1 == 0 && block2 == 0) {
    cWILLY.jp = 0;
    cWILLY.m = 4;
    cCONV.m = 0;
    delayBeep = 1;
  }
  else {
    if (block1 == 5 || block2 == 5 || block1 == 6 || block2 == 6)
      cWILLY.m = 6;
  }
}

/* ////////////////////////////////////////////////////////////
//	Do Willy Fall
//////////////////////////////////////////////////////////// */
static void
DoWillyFall (void)
{
  int block1, block2;

  cWILLY.pos.y += 4;
  cWILLY.jp++;
  if ((SWORD) cWILLY.jp > 11)
    cWILLY.jp = 0;

  block1 = GetBlock (cWILLY.pos.x, cWILLY.pos.y + 16);
  block2 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y + 16);
  if (block1 || block2) {
    cWILLY.m = 0;
    cWILLY.pos.y &= 248;

    cCONV.m = cCONV.m != 0 && cWILLY.fall < 4 ? 1 : 2;
    if (cWILLY.fall >= 32)
      cWILLY.m = 6;
    else
      cWILLY.fall = 0;
  }
  else
    cWILLY.fall += 4;

  if (cWILLY.m == 4) {
    if (delayBeep)
      mm_snd_jump (16384 - (cWILLY.jp * 1000));
    else
      delayBeep=1;
  }
}

/* ////////////////////////////////////////////////////////////
//	Do Willy Normal
//////////////////////////////////////////////////////////// */
static void
PutWilly (void)
{
  if (cWILLY.d) {
    GetWillyBG ((cWILLY.pos.x & 248), cWILLY.pos.y);
    DrawWilly ((cWILLY.pos.x & 248), cWILLY.pos.y, ((cWILLY.pos.x & 15) >> 1) + 8);
  }
  else {
    GetWillyBG ((cWILLY.pos.x & 248), cWILLY.pos.y);
    DrawWilly ((cWILLY.pos.x & 248), cWILLY.pos.y, ((cWILLY.pos.x & 15) >> 1));
  }
}

/* ////////////////////////////////////////////////////////////
//	Get Willy Input
//////////////////////////////////////////////////////////// */
static int
GetWillyInput (void)
{
  int left = 0;
  int right = 0;
  int jump = 0;

  if (mm_keyb_pressed (KEYB_Q) ||
      mm_keyb_pressed (KEYB_E) ||
      mm_keyb_pressed (KEYB_T) ||
      mm_keyb_pressed (KEYB_U) ||
      mm_keyb_pressed (KEYB_O) ||
      mm_keyb_pressed (KEYB_N4) || mm_keyb_pressed (KEYB_OPENSQU)) {
    left = 1;
  }

  if (mm_keyb_pressed (KEYB_W) ||
      mm_keyb_pressed (KEYB_R) ||
      mm_keyb_pressed (KEYB_Y) ||
      mm_keyb_pressed (KEYB_I) ||
      mm_keyb_pressed (KEYB_P) ||
      mm_keyb_pressed (KEYB_N6) || mm_keyb_pressed (KEYB_CLOSEDSQU)) {
    right = 2;
  }

  if (mm_keyb_pressed (KEYB_LEFTSHIFT) ||
      mm_keyb_pressed (KEYB_BACKSLASH) ||
      mm_keyb_pressed (KEYB_Z) ||
      mm_keyb_pressed (KEYB_X) ||
      mm_keyb_pressed (KEYB_C) ||
      mm_keyb_pressed (KEYB_V) ||
      mm_keyb_pressed (KEYB_B) ||
      mm_keyb_pressed (KEYB_N) ||
      mm_keyb_pressed (KEYB_M) ||
      mm_keyb_pressed (KEYB_N8) ||
      mm_keyb_pressed (KEYB_COMMA) ||
      mm_keyb_pressed (KEYB_STOP) ||
      mm_keyb_pressed (KEYB_FWDSLASH) ||
      mm_keyb_pressed (KEYB_RIGHTSHIFT) || mm_keyb_pressed (KEYB_SPACE)) {
    jump = 4;
  }

  return (left & right) ? jump : (left | right | jump);
}

/* ////////////////////////////////////////////////////////////
//	Get Block
//////////////////////////////////////////////////////////// */
static int
GetBlock (WORD x, WORD y)
{
  return (x < 256u && y < 128u) ?
    bMAP[(((y / 8) * 32) + x / 8)] : 3;
}

/* ////////////////////////////////////////////////////////////
//	Check Willy Against Kill Blocks
//////////////////////////////////////////////////////////// */
static int
CheckWillyKillBlock (void)
{
  int block1, block2;
  int block3, block4;
  int block5, block6;

  block1 = GetBlock (cWILLY.pos.x, cWILLY.pos.y);
  block2 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y + 8);
  block3 = GetBlock (cWILLY.pos.x, cWILLY.pos.y + 8);
  block4 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y);
  block5 = GetBlock (cWILLY.pos.x, cWILLY.pos.y + 16);
  block6 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y + 16);

  return
    block1 == 5 || block2 == 5 ||
    block3 == 5 || block4 == 5 ||
    block5 == 5 || block6 == 5 ||
    block1 == 6 || block2 == 6 ||
    block3 == 6 || block4 == 6 || block5 == 6 || block6 == 6;
}

/* ////////////////////////////////////////////////////////////
//	Check Keys
//////////////////////////////////////////////////////////// */
static void
CheckKeys (void)
{
  int i, hit = 0;

  for (i = 4; i >= 0; i--) {
    if ((cKEY[i].s) &&
	cKEY[i].pos.x + 7 >= cWILLY.pos.x && cKEY[i].pos.x < cWILLY.pos.x + 10 &&
	cKEY[i].pos.y + 7 >= cWILLY.pos.y && cKEY[i].pos.y < cWILLY.pos.y + 16) {
      cKEY[i].s = 0;
      hit = 1;
      SCORE += 100;
    }
  }
  if (hit)
    mm_snd_pick (22050);
}

/* ////////////////////////////////////////////////////////////
//	Check Switches
//////////////////////////////////////////////////////////// */
static void
CheckSwitches (void)
{
  int i;

  for (i = 0; i < 2; i++)
    if (cSWITCH[i].s == 1 &&
	cSWITCH[i].pos.x + 7 >= cWILLY.pos.x && cSWITCH[i].pos.y + 7 >= cWILLY.pos.y &&
	cSWITCH[i].pos.x < cWILLY.pos.x + 8 && cSWITCH[i].pos.y < cWILLY.pos.y + 16)
      cSWITCH[i].s = 2;
}

/* ////////////////////////////////////////////////////////////
//	Check Crumb
//////////////////////////////////////////////////////////// */
static void
CheckCrumb (void)
{
  int block1;

  block1 = GetBlock (cWILLY.pos.x, cWILLY.pos.y + 16);
  if (block1 == 4) {
    bCRUMB[(((cWILLY.pos.y + 16) / 8) * 32) + (cWILLY.pos.x / 8)] &= 31;
    bCRUMB[(((cWILLY.pos.y + 16) / 8) * 32) + (cWILLY.pos.x / 8)]--;
    bCRUMB[(((cWILLY.pos.y + 16) / 8) * 32) + (cWILLY.pos.x / 8)] |= 128;

    if (bCRUMB[(((cWILLY.pos.y + 16) / 8) * 32) + (cWILLY.pos.x / 8)] == 128)
      bMAP[(((cWILLY.pos.y + 16) / 8) * 32) + (cWILLY.pos.x / 8)] = 0;
  }

  block1 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y + 16);
  if (block1 == 4) {
    bCRUMB[(((cWILLY.pos.y + 16) / 8) * 32) + ((cWILLY.pos.x + 8) / 8)] &= 31;
    bCRUMB[(((cWILLY.pos.y + 16) / 8) * 32) + ((cWILLY.pos.x + 8) / 8)]--;
    bCRUMB[(((cWILLY.pos.y + 16) / 8) * 32) + ((cWILLY.pos.x + 8) / 8)] |= 128;

    if (bCRUMB[(((cWILLY.pos.y + 16) / 8) * 32) + ((cWILLY.pos.x + 8) / 8)] == 128)
      bMAP[(((cWILLY.pos.y + 16) / 8) * 32) + ((cWILLY.pos.x + 8) / 8)] = 0;
  }
}

/* ////////////////////////////////////////////////////////////
//	Do Crumb
//////////////////////////////////////////////////////////// */
void
DoCrumb (void)
{
  int x, y;

  for (y = 15; y >= 0; --y)
    for (x = 31; x >= 0; --x) {
      BYTE data = bCRUMB[(y * 32) + x];
      BYTE d = data & 31;

      if (data & 192) {
	int dx = x * 8, dy = y * 8;
	if (d < 8)
	  PlotXYback(dx, dy, cBG.paper, 8, 8 - d);

	if (d && d <= 8)
	  PlotXYink (dx, dy + 8 - d, GFXblocks + sCRUMB.gfx * 64,
		     sCRUMB.ink, cBG.paper, 8, d);

	bCRUMB[(y * 32) + x] &= 31;
	if (data & 128)
	  bCRUMB[(y * 32) + x] |= 64;
      }
    }
}

/* ////////////////////////////////////////////////////////////
//	Check Exit
//////////////////////////////////////////////////////////// */
static void
CheckExit (void)
{
  if (cEXIT.m &&
      (signed) cWILLY.pos.x      >= (signed) cEXIT.pos.x -  2 &&
      (signed) cWILLY.pos.x + 10 <= (signed) cEXIT.pos.x + 18 &&
      (signed) cWILLY.pos.y      >= (signed) cEXIT.pos.y -  5 &&
      (signed) cWILLY.pos.y + 16 <= (signed) cEXIT.pos.y + 21)
    GAMEm = 3;

  if (!CHEAT && GAMEm == 3 && LEVEL == 19) {
    GAMEm = 5;
    LASTm = 0;
    LASTc = 0;
    LASTp = 0;
    TEXT.m = 0;
    TEXT.ink = 7;
    TEXT.fade = 0;
  }
}

/* ////////////////////////////////////////////////////////////
//	Check if on CONV
//////////////////////////////////////////////////////////// */
static int
CheckWillyConv (void)
{
  int block1, block2;

  block1 = GetBlock (cWILLY.pos.x, cWILLY.pos.y + 16);
  block2 = GetBlock (cWILLY.pos.x + 8, cWILLY.pos.y + 16);
  return block1 == 7 || block2 == 7;
}

/* ////////////////////////////////////////////////////////////
//	Do Death
//////////////////////////////////////////////////////////// */
static void
DoDeath (void)
{
  GAMEm = 2;
  DEATHm = 0;
  DEATHc = 0;
  mm_snd_stopmod ();
  mm_snd_die (22050);
}

/* ////////////////////////////////////////////////////////////
//	Remove SPG beam
//////////////////////////////////////////////////////////// */
void
RemoveSPG (void)
{
  int i;

  for (i = 0; i < 64; i++) {
    if (SPG[old][i].x < UNDEF)
      DrawBGBlock (SPG[old][i].x, SPG[old][i].y);
    else
      break;
  }
}

/* ////////////////////////////////////////////////////////////
//	Initialise level
//////////////////////////////////////////////////////////// */
void
CopyLevelInfo (void)
{
  int count1;

  memcpy (bMAP, map[LEVEL], sizeof (bMAP));
  for (count1 = 511; count1 >= 0; --count1)
    bCRUMB[count1] = (map[LEVEL][count1] == 4) ? 8 : 0;

  memcpy (bTITLE, title[LEVEL], 32);

  cBG = BG[LEVEL];
  sPLAT1 = PLAT1[LEVEL];
  sPLAT2 = PLAT2[LEVEL];
  sWALL = WALL[LEVEL];
  sCRUMB = CRUMB[LEVEL];
  sKILL1 = KILL1[LEVEL];
  sKILL2 = KILL2[LEVEL];

  sCONV = CONV[LEVEL].s;
  cCONV.pos = CONV[LEVEL].pos;
  cCONV.d = CONV[LEVEL].d;
  cCONV.l = CONV[LEVEL].l;
  cCONV.f = 0;
  cCONV.m = 0;
  cCONV.one = 0;

  cBORDER = BORDER[LEVEL];

  cWILLY.pos = WILLYs[LEVEL];
  cWILLY.d = WILLYsd[LEVEL];
  cWILLY.m = 0;
  cWILLY.f = 0;
  cWILLY.fall = 0;
  cWILLY.j = 0;
  cWILLY.old[0].x = UNDEF;
  cWILLY.old[0].y = UNDEF;
  cWILLY.old[1].x = UNDEF;
  cWILLY.old[1].y = UNDEF;
  cWILLY.jp = 0;

  for (count1 = 0; count1 < 5; count1++) {
    cKEY[count1].pos = KEY[LEVEL].key[count1].pos;
    cKEY[count1].b = (count1 * 2);
    cKEY[count1].s = KEY[LEVEL].key[count1].s;
  }
  cKEYgfx = KEY[LEVEL].gfx;

  cSWITCH[0] = SWITCH[LEVEL][0];
  cSWITCH[1] = SWITCH[LEVEL][1];

  cEXIT.pos = EXIT[LEVEL].pos;
  cEXIT.b = 0;
  cEXIT.m = 0;
  cEXIT.gfx = EXIT[LEVEL].gfx;


  cAIR = AIR[LEVEL] + 31;
  cAIRp = 8;

  for (count1 = 0; count1 < 4; count1++) {
    cHROBO[count1] = HROBO[LEVEL][count1];
    cVROBO[count1] = VROBO[LEVEL][count1];
    cHROBO[count1].old[0].x = UNDEF;
    cHROBO[count1].old[0].y = UNDEF;
    cHROBO[count1].old[1].x = UNDEF;
    cHROBO[count1].old[1].y = UNDEF;
    cVROBO[count1].anim = count1;
    cVROBO[count1].old[0].x = UNDEF;
    cVROBO[count1].old[0].y = UNDEF;
    cVROBO[count1].old[1].x = UNDEF;
    cVROBO[count1].old[1].y = UNDEF;
  }

  EUGENE.pos.x = 120;
  EUGENE.pos.y = 1;
  EUGENE.d = 0;
  EUGENE.m = 0;
  EUGENE.c = 7;
  EUGENE.min = 1;
  EUGENE.max = 87;
  EUGENE.old[0].x = UNDEF;
  EUGENE.old[0].y = UNDEF;
  EUGENE.old[1].x = UNDEF;
  EUGENE.old[1].y = UNDEF;

  SWITCHm[0] = 0;
  SWITCHm[1] = 0;

  HOLEl = 2;
  HOLEy = 95;

  KONG.pos.x = 120;
  KONG.pos.y = 0;
  KONG.max = 104;
  KONG.m = 0;
  KONG.c = 3;
  KONG.f = 0;
  KONG.p = KONGPAUSE;
  KONG.old[0].x = UNDEF;
  KONG.old[0].y = UNDEF;
  KONG.old[1].x = UNDEF;
  KONG.old[1].y = UNDEF;

  SKY[0].p = 0;
  SKY[1].p = 2;
  SKY[2].p = 1;

  SKY[0].s = 4;
  SKY[1].s = 3;
  SKY[2].s = 1;

  SKY[0].c = 6;
  SKY[1].c = 5;
  SKY[2].c = 4;

  SKY[0].max = 72;
  SKY[1].max = 56;
  SKY[2].max = 32;

  for (count1 = 0; count1 < 3; count1++) {
    SKY[count1].pos.x = SKYp[count1][SKY[count1].p].x;
    SKY[count1].pos.y = SKYp[count1][SKY[count1].p].y;
    SKY[count1].m = 0;
    SKY[count1].f = 0;
    SKY[count1].old[0].x = UNDEF;
    SKY[count1].old[0].y = UNDEF;
    SKY[count1].old[1].x = UNDEF;
    SKY[count1].old[1].y = UNDEF;
  }

  for (count1 = 0; count1 < 64; count1++) {
    SPG[0][count1].x = UNDEF;
    SPG[0][count1].y = UNDEF;
    SPG[1][count1].x = UNDEF;
    SPG[1][count1].y = UNDEF;
  }

  DEATHm = 0;
  DEATHc = 0;

  memcpy (PALfade, PALmain, 768);

  SUN.m = 0;
  SUN.y = 32;
  SUN.h = 16;
  SUN.yold[0] = UNDEF;
  SUN.yold[1] = UNDEF;
  SUN.hold[0] = 255;
  SUN.hold[1] = 255;
}
